/*
**  Copyright (C) 1993-1996, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/
/*	
**			ctndisp.c
**
**	CTNDISP is a program for displaying images of arbitrary 
**	size that have been enqueued (using the snq routines) to
**	the Study Name Queue (SNQ).  CTNDISP's principal purpose 
**	is to display images as evidence of a successful
**	transfer of an image by an SCP process.  Images can also
**	be queued via a simple TTY interface (see SNQ) for demonstration
**	and diagnostic purposes.  
**	
**	ACR image names are retrieved from a shared memory segment that
**	is created by the SCP process.  This shared memory segment,
**	known as the Study Name Queue (SNQ), is a FIFO queue
**	on which the IDX image database names are enqueued.
**
**	CTNDISP always displays the first image of any given study.
**	Images are displayed using gray-scale lookup table.  The patient
**	name, date, modality, and study are displayed above each image.  The
**	node name is displayed at the top.  
**
**	 
** $Log: ctndisp.cc,v $
** Revision 1.7  1997/10/22 09:21:06  meichel
** Corrected check for presence of shared memory system calls in ctndisp.
**   Needed for AIX 3.2 (IBM xlC 1.0) where shmctl and shmdt are present, but shmat is not.
**
** Revision 1.6  1997/07/24 13:10:42  andreas
** - Removed Warnings from SUN CC 2.0.1
**
** Revision 1.5  1997/05/30 12:29:30  andreas
** - include version number in every usage String
**
** Revision 1.4  1997/05/29 15:52:47  meichel
** Added constant for dcmtk release date in dcuid.h.
** All dcmtk applications now contain a version string
** which is displayed with the command line options ("usage" message)
** and which can be queried in the binary with the "ident" command.
**
** Revision 1.3  1997/05/22 13:23:41  hewett
** Modified the test for presence of a data dictionary to use the
** method DcmDataDictionary::isDictionaryLoaded().
**
** Revision 1.2  1996/04/18 15:29:40  meichel
** Version for CAR'96
**
** Revision 1.1  1996/03/28 17:06:36  meichel
** first commit with new file structure
**
** Revision 2.2  1995/05/18 11:57:01  hewett
** Ran code through a formatter.  I now a little more readable.
**
 * Revision 2.1  1995/02/02  12:49:25  hewett
 * Corrected spelling of Ossietzky
 *
 * Revision 2.0  1994/11/17  09:44:08  meichel
 * Version 2.0 for RSNA 1994 demonstration.
 * New features: Metaheader support, new SOP classes, lots of bugs fixed.
 *
 * Revision 1.15  1994/03/17  09:27:46  hewett
 * Updated copyright notice.
 *
 * Revision 1.14  1994/03/16  16:17:27  hewett
 * Reduced the number of gray levels.  Get really wierd results on
 * Solaris2 otherwise.
 *
 * Revision 1.13  1993/11/22  12:31:49  hewett
 * Removed 10mb queue input buffer.  No longer necessary since snq fixed.
 *
 * Revision 1.12  1993/11/19  15:11:33  jens
 * Fixed some bugs.
 *
 * Revision 1.11  1993/11/18  20:17:01  jens
 * Multi processes Version of ctndisp.
 * If you don`t want forked image conversion, then use -noMulti switch
 *
 * Revision 1.10  1993/11/16  12:24:30  jens
 * Added switch -2images to ctndisp. Only 2 Imageareas are used.
 * Modified the Assiciation window refresh. (less flicker)
 * Added a image count information.
 * Fasten XputImage.
 * Added code for the killassoc command.
 *
 * Revision 1.9  1993/11/10  17:42:55  jens
 * Added a new flag: -half
 * with ctndisp -half the window uses a square half of the screen
 *
 * Revision 1.8  1993/11/01  21:02:16  jens
 * Better error resistent Version in Tag reading section.
 *
 * Revision 1.7  1993/10/14  10:46:08  hewett
 * Added Euro Header
 *
 * Revision 1.6  1993/10/14  10:16:59  hewett
 * Cleaned up header.
 *
**
** $Id: ctndisp.cc,v 1.7 1997/10/22 09:21:06 meichel Exp $
**		
*/

#include "osconfig.h" /* configure mechanism */
#include "dcuid.h"    /* for dcmtk version name */

static char rcsid[] = "$dcmtk: ctndisp v" 
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";

#include <stdio.h>
#include <stdlib.h>

#ifndef X_DISPLAY_MISSING
#if defined(HAVE_SHMAT) && defined(HAVE_SHMCTL) && defined(HAVE_SHMGET)

#include <string.h>
#include <math.h>

extern "C" {
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#ifdef HAVE_LIBC_H
#include <libc.h>
#endif

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_IPC_H
#include <sys/ipc.h>
#endif

#ifdef HAVE_SYS_SHM_H
#include <sys/shm.h>
#else
/* 
   This is a dummy declaration of struct shmid_ds. The real declaration from 
   sys/shm.h is very complex. Since we only pass "opaque" pointers of this 
   type to shmctl() and don't need the results, this hack should work.
   Anyway: Which weard kind of system
   should support shmctl() and have no appropriate header file??? 
 */
     struct shmid_ds { int dummy_element; };
#endif

#ifndef HAVE_PROTOTYPE_SHMCTL
     int shmctl(int shmid, int cmd, struct shmid_ds *buf);
#endif

#ifdef HAVE_USLEEP
#ifndef HAVE_PROTOTYPE_USLEEP
     int usleep(unsigned int useconds);
#endif
#endif

}

#include "snq.h"
#include "disppriv.h"
#include "ctndsupp.h"
#include "ctnddico.h"
#include "dcdict.h"

static void usage(FILE *f)
{
  fprintf(f, "%s\n\nusage: ctndisp [-n node  -w display_width"
    " -h display_height -half -2images -noRescale -noWindow -noMulti]\n",
    rcsid);
  return;
}

/*
 * This typedef and these function prototypes occur here because gcc can't
 * handlel the X include libraries....this is the short answer...
 */
typedef struct _XImageStruct {
    XImage *x_image;		/* the X image struct		 */
    Window image_window;	/* the window for display	 */
}   XImageStruct;

void 
CTNDISP_DisplayImageData(ImageStruct * img, XImageStruct * ximg);
void CTNDISP_RefreshImageData();

/*
 * Globals (preceded with G_)...declared extern in other files
 */
Window G_big_window;

Display *G_display;		/* Xstuff used all over...	 */

Visual *G_visual;

Colormap G_cmap;

GC  G_gclargeT,
    G_gcsmallT,
    G_gcsmallTbold,
    G_gcblack,			/* Graphics contexts needed...	 */
    G_gcred,
    G_gcgreen;

int G_display_width,
    G_display_height,		/* Well know screen stuff...	 */
    G_screen,
    G_ncolors;

int G_FirstTime = 1,
    Rescaleflag = 1,
    Windowflag = 1;		/* if to use the rescale and window data.
				 * default is yes */
char G_nodename[MAX_STRING_LENGTH];

QUEUE_Element Ctndisplayhandle,
    CalcElement;		/* CalcElement is the Image element in
				 * calculation (if MULTI) */

int MULTI = 1;

int ImageQueueStatus = EMPTY;
int CalcStatus = IFREE;


unsigned char *ishmaddr,
   *tishmaddr;			/* shared memory adress pointer */
int status,
    died,
    pid,
    shmflag;			/* status of son */
struct shmid_ds *shm_buf;
key_t ishmkey;
int ishmsize,
    ishmflag,
    ishmID;


/*
 * End Globals...
 */

ImageStruct cur_img,
    next_img;

XImageStruct xcur_img,
    xnext_img;

Image_Modul tempimage;


/*---------------------*/
void
InitSemaphore(int study_Q_id)
{

    unsigned long retval;

#ifdef DEBUG
    printf("Init Semaphore\n");
#endif
    /*
     * Initialize the SNQ for studies and status messages 
     */

    retval = SNQ_InitQueue(study_Q_id, QUEUE_SIZE);
#ifdef DEBUG
    printf("SNQ_InitQueue return value = %ld\n", retval);
#endif

    if (retval != SNQ_NORMAL) {
	fprintf(stderr,"CTNDISP: Not able to access Study-Name Queue.\n");
	exit(-1);
    }
} /* end InitSemaphore */


unsigned long
GetElement(QUEUE_Element * tempelement)
{
    return SNQ_Dequeue(tempelement);
}


/*******************************************************************/
main(int argc, char **argv)
{
    unsigned long retval;
    QUEUE_Element element;
    int count_args,
        node_arg,
        study_Q_id;

    XEvent event;

    memset(&tempimage, 0, sizeof(Image_Modul));

    /*
     * Get the input paramaters from command line 
     */
    if (argc < 1) {
        usage(stderr);
	exit(-1);
    }
    study_Q_id = -1;

    G_display_width = G_display_height = -1;
    /* -1's indicate to set the width and height to full screen */

    count_args = node_arg = 0;
    argc--;
    argv++;
    while (argc > 0) {
	if ((strcmp(*argv, "-q")) == 0) {
	    argc--;
	    argv++;
	    study_Q_id = atoi(*argv);
	    if (study_Q_id < 0) {
		fprintf(stderr,  "CTNDISP: Queue number [%d] must > 0\n", study_Q_id);
		exit(-1);
	    }
	    argc--;
	    argv++;
	} else if ((strcmp(*argv, "-half")) == 0) {
	    argc--;
	    argv++;
	    G_display_height = -2;
	    G_display_width = -2;	/* indicates in initXwindows that to
					 * use the half screen height and
					 * width */

	    count_args++;
	} else if ((strcmp(*argv, "-2images")) == 0) {
	    argc--;
	    argv++;
	    G_display_height = -3;	/* indicates in initXwindows that to
					 * use the half screen height */

	    Set_Images_and_Show(2, 10);
	    count_args++;
	} else if ((strcmp(*argv, "-w")) == 0) {
	    argc--;
	    argv++;
	    G_display_width = atoi(*argv);
	    if (G_display_width < MIN_DISPLAY_WIDTH) {
		fprintf(stderr,  "CTNDISP: Display width must be > MIN_DISPLAY_WIDTH [%d]\n", MIN_DISPLAY_WIDTH);
		G_display_width = -1;
	    }
	    count_args++;
	    argc--;
	    argv++;
	} else if ((strcmp(*argv, "-h")) == 0) {
	    argc--;
	    argv++;
	    G_display_height = atoi(*argv);
	    if (G_display_height < MIN_DISPLAY_HEIGHT) {
		fprintf(stderr,  "CTNDISP: Display height must be > MIN_DISPLAY_HEIGHT [%d]\n", MIN_DISPLAY_HEIGHT);
		G_display_height = -1;
	    }
	    count_args++;
	    argc--;
	    argv++;
	} else if ((strcmp(*argv, "-n")) == 0) {
	    argc--;
	    argv++;
	    strcpy(G_nodename, *argv);
	    node_arg++;
	    argc--;
	    argv++;
	} else if ((strcmp(*argv, "-noRescale")) == 0) {
	    argc--;
	    argv++;
	    Rescaleflag = 0;
	    printf("Rescale Intercept/Slope Information ignored !\n");
	} else if ((strcmp(*argv, "-noWindow")) == 0) {
	    argc--;
	    argv++;
	    Windowflag = 0;
	    printf("Window Center/Width Information ignored !\n");

	} else if ((strcmp(*argv, "-noMulti")) == 0) {
	    argc--;
	    argv++;
	    MULTI = 0;
	    printf("No Multiprocessing !\n");
	} else {
            usage(stderr);
	    exit(-1);
	}
    }
    if (node_arg == 0)
	strcpy(G_nodename, "UNKNOWN");

    if (study_Q_id < 0) {
	/*
	 * printf("Argument Queue Number missing: -q Number\nSet to 0\n"); 
	 */
	study_Q_id = 0;
    }

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded()) {
	fprintf(stderr, "Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }

    /*
     * Initialize the SNQ for studies and status messages 
     */
    InitSemaphore(study_Q_id);

    /*
     * Initialize the X stuff, load gray-scale color table 
     */
    if (! CTNDISP_InitializeX(220)) {
	SNQ_KillQueue(study_Q_id);
        /* error msg already printed from CTNDISP_InitializeX */
	exit(-1);
    }
    /*
     * Display node name and the status buttons... 
     */
    CTNDISP_CreateShow();	/* do all the init stuff */


    /*
     * Main Control loop for ctndisp 
     *
     */

    /* NEW STUFF J.Burkert */

    while (1) {

	/* printf("%d events pending\n",XPending(G_display)); */

	if (XPending(G_display) > 0) {	/* first look if there is a xwindow
					 * event, if not examine the QUEUE
					 * else sleep a while */
	    XNextEvent(G_display, &event);

	    /* printf("event number : %d\n",event.type); */
	    switch (event.type) {
	    case ConfigureNotify:
		/* printf("Config notify\n"); */
		break;
	    case Expose:
		if (event.xexpose.count != 0)
		    break;	/* to avoid redraw often in a row */
		CTNDISP_RefreshImageData(event);

		CTNDISP_RefreshStatus();
		break;
	    case UnmapNotify:

	    case DestroyNotify:
		if (event.xdestroywindow.window == G_big_window) {
		    if ((SNQ_KillQueue(study_Q_id)) != SNQ_NORMAL) 
			fprintf(stderr, "CTNDISP: Not able to kill Study-Name Queue.\n");
		    if (MULTI) {
			if ((shmctl(ishmID, IPC_RMID, shm_buf)) == -1) {
			    shared_error();
			}
		    }
		    printf("CTNDISPLAY terminated. Bye \n");
		    exit(0);
		}
		break;
	    default:
		break;

	    } /* switch */
	}
	/* if Pending */
	else {
	    retval = GetElement(&element);
	    if (retval == SNQ_NORMAL) {


#ifdef DEBUG
		printf("show it \n");
#endif
		ShowElement(element);

		/* CTNDISP_RefreshStatus(); */
	    } else if (MULTI) {
		if (CalcStatus == ICALC) {
		    status = 0;
		    died = (int)waitpid(pid, &status, WNOHANG | WUNTRACED);
		    if ((status >> 8) == 42) {
			died = (int)wait(&status);

#ifdef DEBUG
			printf("IREADY in ctndisp.c \n");
			printf("%d WIFEXITED\n", WIFEXITED(status));
			printf("%d WEXITSTATUS\n", WEXITSTATUS(status));
			printf("%d WSTOPSIG\n", WSTOPSIG(status));
#endif

			CalcStatus = IREADY;
			ShowElement(CalcElement);
			CalcStatus = IFREE;
		    }
		} /* CALC */
		if ((CalcStatus == IFREE) && (ImageQueueStatus == NOTEMPTY)) {
#ifdef DEBUG
		    printf("IFREE in ctndisp.c \n");
#endif

		    CalcStatus = INEXT;
		    ImageQueue(&CalcElement, GET);	/* get image from queue */

		    ShowElement(CalcElement);
#ifdef DEBUG
		    printf("ICALC in ctndisp.c \n");
#endif

		    /*
		     * Calctstatus = icalc is set in ShowImage 
		     */
		} else {	/* default */
#ifdef HAVE_USLEEP
		    usleep(10000);	/* sleep 0.01 sec */
#else
		    sleep(1);
#endif
		} /* if there are no importent data queued */
	    } else {		/* if MULTI */
#ifdef HAVE_USLEEP
		usleep(10000);	/* sleep 0.01 sec */
#else
		sleep(1);
#endif
	    }

	} /* else */
    } /* while */
}

#else /* not HAVE_SHMxxx */

main()
{
    fprintf(stderr, 
      "%s\n\n"
      "ctndisp requires shared memory (system calls: shmat, shmctl, shmget).\n"
      "configure detected that your system does not support shared memory.\n"
      "Please reconfigure your system and recompile if appropriate.\n",
      rcsid);
}

#endif
#else /* X_DISPLAY_MISSING */

main()
{
    fprintf(stderr, 
      "%s\n\n"
      "ctndisp requires the X Windows System.\n"
      "configure detected that your system does not support X11.\n"
      "Please reconfigure your system and recompile if appropriate.\n",
      rcsid);
}

#endif /* X_DISPLAY_MISSING */
