/*
**  Copyright (C) 1993/1994, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/
/*
** Service Class Executive (SCE)
** - Store Service Class Provider
**
** Author: 	Andrew Hewett
**		Universitaet Oldenburg, Germany
** Created:	08/1993
**
** Last Update:		$Author: hewett $
** Update Date:		$Date: 1998/01/14 14:29:47 $
** Source File:		$Source: /share/dicom/cvs-depot/dcmtk/imagectn/apps/scestore.cc,v $
** CVS/RCS Revision:	$Revision: 1.10 $
** Status:		$State: Exp $
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
END_EXTERN_C

#include "dicom.h"
#include "imagectn.h"
#include "dimse.h"
#include "diutil.h"
#include "sce.h"
#include "imagedb.h"
#include "cnf.h"
#ifndef NODISPLAY
#include "dispuser.h"
#endif
#include "dcfilefo.h"

#include "scestore.h"

/*
 * C-STORE SCP
 */


typedef struct {
    DB_Handle *dbHandle;
    DIC_US status;
    DcmDataset *statusDetail;
    char *fileName;
    DcmFileFormat *dcmff;
} SCE_StoreContext;


static void
updateDisplay(T_DIMSE_StoreProgress * progress)
{
    int percent;
    static int lastPercent = 0;

    switch (progress->state) {
    case DIMSE_StoreBegin:
	if (verbose) printf("RECV:");
	lastPercent = 0;
	break;
    case DIMSE_StoreEnd:
	if (verbose) printf("\n");
	if (displayTransferProgress) {
#ifndef NODISPLAY
	    DSU_operationState(DSU_MainAssociation, DSU_StoreOperation,
		DSU_RunningOperation, DSU_Incoming, 100);
#endif
	}
	break;
    default:
	if (verbose) putchar('.');
	if (progress->totalBytes == 0) {
	    percent = 100;
	} else {
	    percent = (int)(((float)progress->progressBytes / 
	        (float)progress->totalBytes) * 100.0);
	}
	if (displayTransferProgress) {
	    if ((percent - lastPercent) >= 10) {	/* 10% change */
		if (lastPercent < 100) {
#ifndef NODISPLAY
	            DSU_operationState(DSU_MainAssociation, DSU_StoreOperation,
		        DSU_RunningOperation, DSU_Incoming, percent);
#endif
		}
	        lastPercent = percent;
	    }
	}
	break;
    }
    if (verbose) fflush(stdout);
}


static void 
saveImageToDB(
    SCE_StoreContext *context,
    T_DIMSE_C_StoreRQ *req,		/* original store request */
    char *imageFileName,
    /* out */
    T_DIMSE_C_StoreRSP *rsp, 		/* final store response */
    DcmDataset **statusDetail)
{
    CONDITION dbcond;
    DB_Status dbStatus;

    dbStatus.status = STATUS_Success;
    dbStatus.statusDetail = NULL;
    
    /* Store image */
    if (ignoreStoreData) {
        rsp->Status = STATUS_Success;
	*statusDetail = NULL;
	return;	/* nothing else to do */
    }
    
    if (context->status == STATUS_Success) {

	if (dbStatus.status == STATUS_Success) {
    
            dbcond = DB_storeRequest(context->dbHandle,
	        req->AffectedSOPClassUID, req->AffectedSOPInstanceUID,
	        imageFileName, &dbStatus);
            if (dbcond != DB_NORMAL) {
                errmsg("storeSCP: Database: DB_storeRequest Failed (%s)",
	           DU_cstoreStatusString(dbStatus.status));
	        COND_DumpConditions();
	    }
	}
    } else {
        dbStatus.status = context->status;
	dbStatus.statusDetail = context->statusDetail;
    }
    rsp->Status = dbStatus.status;
    *statusDetail = dbStatus.statusDetail;
    context->status = dbStatus.status;
    context->statusDetail = dbStatus.statusDetail;
}

static void
writeToFile(
    DcmFileFormat *ff,
    const char* fileName,
    T_DIMSE_C_StoreRSP *rsp)
{

    DcmFileStream outf( fileName, DCM_WriteMode );
    if ( outf.Fail() ) {
	errmsg("Cannot write image file: %s", fileName);
	rsp->Status = STATUS_STORE_Refused_OutOfResources;
    } else {

	E_TransferSyntax xfer = EXS_Unknown;
	E_EncodingType enctype = EET_ExplicitLength;
	E_GrpLenEncoding ogltype = EGL_recalcGL;

	if (xfer == EXS_Unknown) {
	    /* use the same as the input */
	    xfer = ff->getDataset()->getOriginalXfer();
	}

	ff->transferInit();
	ff->write( outf, xfer, enctype, ogltype );
	ff->transferEnd();

	if (ff->error() != EC_Normal) {
	    errmsg("Cannot write image file: %s", fileName);
	    rsp->Status = STATUS_STORE_Refused_OutOfResources;
	}
    }
}

static void
checkRequestAgainstDataset(
    T_DIMSE_C_StoreRQ *req,     /* original store request */
    const char* fname,          /* filename of dataset */
    DcmDataset *dataSet,        /* dataset to check */
    T_DIMSE_C_StoreRSP *rsp) 	/* final store response */
{
    DcmFileFormat ff;

    if (dataSet == NULL) {
	/* load the data from file */
	DcmFileStream istrm(fname, DCM_ReadMode);
	if ( istrm.Fail() ) {
	    errmsg("Cannot open file: %s", fname);
	    rsp->Status = STATUS_STORE_Refused_OutOfResources;
	    return;
	}

	ff.read(istrm, EXS_Unknown, EGL_noChange );

	dataSet = ff.getDataset();
    }

    /* which SOP class and SOP instance ? */
    DIC_UI sopClass;
    DIC_UI sopInstance;
    
    if (!DU_findSOPClassAndInstanceInDataSet(dataSet, sopClass, sopInstance)) {
	errmsg("Bad image file: %s", fname);
	rsp->Status = STATUS_STORE_Error_CannotUnderstand;
    } else if (strcmp(sopClass, req->AffectedSOPClassUID) != 0) {
	rsp->Status = STATUS_STORE_Error_DataSetDoesNotMatchSOPClass;
    } else if (strcmp(sopInstance, req->AffectedSOPInstanceUID) != 0) {
	rsp->Status = STATUS_STORE_Error_DataSetDoesNotMatchSOPClass;
    }
}

static void
storeProgressCallback(
    /* in */
    void *callbackData, 
    T_DIMSE_StoreProgress *progress,	/* progress state */
    T_DIMSE_C_StoreRQ *req,		/* original store request */
    char *imageFileName,       /* being received into */ 
    DcmDataset **imageDataSet, /* being received into */
    /* out */
    T_DIMSE_C_StoreRSP *rsp, 		/* final store response */
    DcmDataset **statusDetail)
{
    SCE_StoreContext *context;

    updateDisplay(progress);

    if (progress->state == DIMSE_StoreEnd) {
        context = (SCE_StoreContext*)callbackData;	/* recover context */

	if (!ignoreStoreData && rsp->Status == STATUS_Success) {
	    if ((imageDataSet)&&(*imageDataSet)) {
		checkRequestAgainstDataset(req, NULL, *imageDataSet, rsp);
	    } else {
		checkRequestAgainstDataset(req, imageFileName, NULL, rsp);
	    }
	}

	if (!ignoreStoreData && rsp->Status == STATUS_Success) {
	    if ((imageDataSet)&&(*imageDataSet)) {
		writeToFile(context->dcmff, context->fileName, rsp);
	    }

	    saveImageToDB(context, req, context->fileName, rsp, statusDetail);
	}
    }
}

CONDITION
SCE_storeSCP(T_ASC_Association * assoc, T_DIMSE_C_StoreRQ * request,
	     T_ASC_PresentationContextID presId,
	     DB_Handle *dbHandle)
{
    CONDITION cond = DIMSE_NORMAL;
    CONDITION dbcond = DB_NORMAL;
    char imageFileName[MAXPATHLEN+1];
    static OFBool firstImageOnAssociation = OFTrue;
    DIC_AE peerAETitle;
    DIC_AE myAETitle;
    const char *vendorName;
    char displayString[256];
    SCE_StoreContext context;
    int lockfd;
    DcmFileFormat dcmff;

    context.dbHandle = dbHandle;
    context.status = STATUS_Success;
    context.statusDetail = NULL;
    context.dcmff = &dcmff;

    if (verbose) {
	printf("Received Store SCP: ");
        DIMSE_printCStoreRQ(stdout, request);
    }

    if (!dcmIsaStorageSOPClassUID(request->AffectedSOPClassUID)) {
	/* callback will send back sop class not supported status */ 
	context.status = STATUS_STORE_Refused_SOPClassNotSupported;
	/* must still receive data */
	strcpy(imageFileName, "/dev/null");
    } else if (ignoreStoreData) {
	strcpy(imageFileName, "/dev/null");
    } else {
	dbcond = DB_makeNewStoreFileName(dbHandle,
	    request->AffectedSOPClassUID,
	    request->AffectedSOPInstanceUID,
	    imageFileName);
	if (dbcond != DB_NORMAL) {
	    errmsg("storeSCP: Database: DB_makeNewStoreFileName Failed");
	    /* must still receive data */
	    strcpy(imageFileName, "/dev/null");	
	    /* callback will send back out of resources status */ 
	    context.status = STATUS_STORE_Refused_OutOfResources;
	}
    }

    /* exclusively lock image file */
#ifdef O_BINARY
    lockfd = open(imageFileName, (O_WRONLY | O_CREAT | O_TRUNC | O_BINARY), 0666);
#else
    lockfd = open(imageFileName, (O_WRONLY | O_CREAT | O_TRUNC), 0666);
#endif
    flock(lockfd, LOCK_EX);

    context.fileName = imageFileName;

    DcmDataset *dset = dcmff.getDataset();

    /* we must still retrieve the data set even if some error has occured */

    if (opt_bypass) { /* the bypass option can be set on the command line */
	cond = DIMSE_storeProvider(assoc, presId, request, imageFileName, 1,
				   NULL, storeProgressCallback, 
				   (void*)&context, DIMSE_BLOCKING, 0);
    } else {
	cond = DIMSE_storeProvider(assoc, presId, request, (char *)NULL, 1,
				   &dset, storeProgressCallback, 
				   (void*)&context, DIMSE_BLOCKING, 0);
    }


    if (!SUCCESS(cond)) {
        errmsg("Store SCP Failed:");
	COND_DumpConditions();
    }
    if (!ignoreStoreData && 
        (!SUCCESS(cond) || (context.status != STATUS_Success))) {
        /* remove file */
        if (verbose) {
	    fprintf(stderr, "Store SCP: Deleting Image File: %s\n", 
		    imageFileName);
        }
        unlink(imageFileName);
	DB_pruneInvalidRecords(dbHandle);
    }

    /* unlock image file */
    flock(lockfd, LOCK_UN);
    close(lockfd);
    
    if (firstImageOnAssociation && (context.status == STATUS_Success) && 
        !ignoreStoreData && 
	peerCanDisplayImages(assoc->params->DULparams.callingPresentationAddress)) {
	/*
	 * This works since each association runs in a separate
	 * process and firstImageOnAssociation is a static variable
	 * initialised to OFTrue.
	 */
        ASC_getAPTitles(assoc->params, peerAETitle, myAETitle, NULL);
        vendorName = CNF_vendorForPeerAETitle(peerAETitle);
	sprintf(displayString, "%s -> %s", 
	    ((vendorName!=NULL)?(vendorName):(peerAETitle)),
	    myAETitle);
#ifndef NODISPLAY
	DSU_displayIndependent(displayString, imageFileName);
#endif
	firstImageOnAssociation = OFFalse;
    }

    return cond;
}

/*
** CVS Log
** $Log: scestore.cc,v $
** Revision 1.10  1998/01/14 14:29:47  hewett
** Added basic support for the Structured Reporting (SR) SOP Classes.
**
** Revision 1.9  1997/08/26 14:17:20  hewett
** Added +B command line option to imagectn application.  Use of this option
** causes imagectn to bypass the dcmdata encode/decode routines when receiving
** images and write image data to disk exactly as received in a C-STORE
** command over the network.  This option does _not_ affect sending images.
**
** Revision 1.8  1997/08/06 12:20:17  andreas
** - Using Windows NT with Visual C++ 4.x the standard open mode for files
**   is TEXT with conversions. For binary files (image files, imagectn database
**   index) this must be changed (e.g. fopen(filename, "...b"); or
**   open(filename, ..... |O_BINARY);)
**
** Revision 1.7  1997/07/21 08:59:45  andreas
** - Replace all boolean types (BOOLEAN, CTNBOOLEAN, DICOM_BOOL, BOOL)
**   with one unique boolean type OFBool.
**
** Revision 1.6  1997/06/26 12:59:20  andreas
** - Include Additional headers (winsock.h, io.h) for Windows NT/95
**
** Revision 1.5  1997/05/23 10:47:08  meichel
** Major rewrite of storescp application. See CHANGES for details.
** Changes required to interfaces of some DIMSE functions.
**
** Revision 1.4  1997/05/16 08:31:41  andreas
** - Revised handling of GroupLength elements and support of
**   DataSetTrailingPadding elements. The enumeratio E_GrpLenEncoding
**   got additional enumeration values (for a description see dctypes.h).
**   addGroupLength and removeGroupLength methods are replaced by
**   computeGroupLengthAndPadding. To support Padding, the parameters of
**   element and sequence write functions changed.
**
** Revision 1.3  1996/09/27 08:46:22  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C.
**
** Revision 1.2  1996/04/22 10:26:25  hewett
** Now calls DB_pruneInvalidRecords on error.
**
** Revision 1.1.1.1  1996/03/28 19:24:59  hewett
** Oldenburg Image CTN Software ported to use the dcmdata C++ toolkit.
**
**
*/
